from collections import deque
from typing import Deque

from cfn_check.shared.types import (
    Data,
    Items,
    YamlObject,
)

from .parsing import QueryParser

class Evaluator:

    def __init__(self):
        self._query_parser = QueryParser()

    def match(
        self,
        resources: YamlObject,
        path: str,
    ):
        items: Items = deque()
        items.append(resources)

        segments = path.split("::")[::-1]
        # Queries can be multi-segment,
        # so we effectively perform per-segment
        # repeated DFS searches, returning the matches
        # for each segment

        composite_keys: list[str] = []

        while len(segments):
            query = segments.pop()
            items, keys = self._match_with_query(items, query)

            if len(composite_keys) == 0:
                composite_keys.extend(keys)

            else:
                updated_keys: list[str] = []
                for composite_key in composite_keys:
                    while len(keys):
                        key = keys.pop()

                        updated_keys.append(f'{composite_key}.{key}')

                composite_keys = updated_keys

        assert len(composite_keys) == len(items), f'❌ {len(items)} returned for {len(composite_keys)} keys. Are you sure you used a range ([*]) selector?'

        results: list[tuple[str, Data]] = []
        for idx, item in enumerate(list(items)):
            results.append((
                composite_keys[idx],
                item,
            ))

        return results
    
    def _match_with_query(
        self,
        items: Items,
        query: str,
    ) -> tuple[Items, Deque[str]]:
        
        found: Items = deque()
        keys: Deque[str] = deque()

        tokens = self._query_parser.parse(query)
        
        while len(items):
            node = items.pop()

            for token in tokens:
                matched_keys, matches = token.match(node)

                if matched_keys and matches:
                    keys.extend(matched_keys)
                    found.extend(matches)

                elif matched_keys is None and matches:
                    items.extend(matches)
            
        return found, keys



from typing import TypeVar
from pydantic import BaseModel, JsonValue
from typing import Callable
from cfn_check.validation.validator import Validator


T = TypeVar("T", bound= JsonValue | BaseModel)


class Rule:

    def __init__(
        self,
        query: str,
        name: str,
    ):
        self.query = query
        self.name = name

    def __call__(self, func: Callable[[T], None]):
        return Validator[T](
            func,
            self.query,
            self.name,
        )



from enum import Enum


class TokenType(Enum):
    BOUND_RANGE = "BOUND_RANGE"
    INDEX = "INDEX"
    KEY = "KEY"
    KEY_RANGE = "KEY_RANGE"
    NESTED_RANGE = "NESTED"
    PATTERN = "GLOB"
    PATTERN_RANGE = "GLOB_RANGE"
    UNBOUND_RANGE = "UNBOUND_RANGE"
    WILDCARD = "WILDCARD"
    WILDCARD_RANGE = "WILDCARD_RANGE"




import asyncio
import os
import pathlib
import yaml
from cfn_check.loader.loader import (
    Loader,
    create_tag,
    find_templates,
)
from cfn_check.shared.types import YamlObject, Data

def open_template(path: str) -> YamlObject | None:

    if os.path.exists(path) is False:
        return None

    try:
        with open(path, 'r') as f:
            return yaml.load(f, Loader=Loader)
    except Exception as e:
        raise e
    
def is_file(path: str) -> bool:
    return os.path.isdir(path) is False


async def path_exists(path: str, loop: asyncio.AbstractEventLoop):
    return await loop.run_in_executor(
        None,
        os.path.exists,
        path,
    )

async def convert_to_cwd(loop: asyncio.AbstractEventLoop):
    return await loop.run_in_executor(
        None,
        os.getcwd,
    )

async def localize_path(path: str, loop: asyncio.AbstractEventLoop):
    localized = path.replace('~/', '')

    home_directory = await loop.run_in_executor(
        None,
        pathlib.Path.home,
    )

    return await loop.run_in_executor(
        None,
        os.path.join,
        home_directory,
        localized,
    )

async def load_templates(
    path: str,
    tags: list[str],
    file_pattern: str | None = None,
):

    loop = asyncio.get_event_loop()
    
    if path == '.':
        path = await convert_to_cwd(loop)

    elif path.startswith('~/'):
        path = await localize_path(path, loop)

    if await loop.run_in_executor(
        None,
        is_file,
        path,
    ) or file_pattern is None:
        template_filepaths = [
            path,
        ]

        assert await path_exists(path, loop) is True, f'❌ Template at {path} does not exist'

    elif file_pattern:

        template_filepaths = await loop.run_in_executor(
            None,
            find_templates,
            path,
            file_pattern,
        )

    assert len(template_filepaths) > 0 , '❌ No matching files found'
    
    for tag in tags:
        new_tag = await loop.run_in_executor(
            None,
            create_tag,
            tag,
        )

        Loader.add_constructor(f'!{tag}', new_tag)

    
    templates: list[Data]  = await asyncio.gather(*[
        loop.run_in_executor(
            None,
            open_template,
            template_path,
        ) for template_path in template_filepaths
    ])

    found_templates = [
        template for template in templates if template is not None
    ]

    assert len(found_templates) > 0, "❌ Could not open any templates"

    return templates



from pydantic import ValidationError

from cfn_check.validation.validator import Validator
from .errors import assemble_validation_error
from .evaluator import Evaluator

class ValidationSet:

    def __init__(
        self,
        validators: list[Validator],
    ):
        self._evaluator = Evaluator()
        self._validators = validators

    @property
    def count(self):
        return len(self._validators)

    def validate(
        self,
        templates: list[str],
    ):
        errors: list[Exception | ValidationError] = []

        for template in templates:
            for validator in self._validators:
                if errs := self._match_validator(
                    validator,
                    template,
                ):
                    errors.extend([
                        (
                            validator,
                            err
                        ) for err in errs
                    ])

        if validation_error := assemble_validation_error(errors):
            return validation_error 

    def _match_validator(
        self,
        validator: Validator,
        template: str,
    ):
        found = self._evaluator.match(template, validator.query)

        assert len(found) > 0, f"❌ No results matching results for query {validator.query}"

        errors: list[Exception | ValidationError] = []


        for matched in found:
            if err := validator(matched):
                errors.append(err)

        if len(errors) > 0:
            return errors
        